#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "bresnham.h"
#include "hp2xx.h"
#include "hpgl.h"
#include "lindef.h"
#include "pendef.h"
#include <assert.h>

/* Calculate and return the centroid of a polygon.
   Assumes poly[points] = poly[0].
*/
static HPGL_Pt centroid(const HPGL_Pt poly[], int points)
{
    HPGL_Pt result;
    double  wsum = 0.0;
    double  xsum = 0.0;
    double  ysum = 0.0;
    int     i;

    for (i = 0; i < points; i++) {
        const double  w = (poly[i].x*poly[i+1].y) - (poly[i].y*poly[i+1].x);
        wsum += w;
        xsum += w * (poly[i].x + poly[i+1].x);
        ysum += w * (poly[i].y + poly[i+1].y);
    }
    
    wsum += wsum + wsum;  /* wsum = 3*wsum; */

    if (wsum != 0.0) {
        result.x = xsum / wsum;
        result.y = ysum / wsum;
    } else {
        result = poly[0];
    }

    return result;
}

/*
 * Polygon rotation and scaling
*/

/* Rotating by angle a, scaling by factor s, with (x0,y0) moved to origin:
    new_x = (old_x - x0)*cos(a)*s + (old_y - y0)*sin(a)*s
          = old_x * ( cos(a)*s )
          + old_y * ( sin(a)*s )
          +         ( - x0 * cos(a)*s - y0 * sin(a)*s )
    new_y = (old_y - y0)*cos(a)*s - (old_x - x0)*sin(a)*s
          = old_x * ( -sin(a)*s )
          + old_y * (  cos(a)*s )
          +         ( x0 * sin(a)*s - y0 * cos(a)*s )
   Inversely:
    old_x = x0 + new_x*cos(a)/s - new_y*sin(a)/s
          = new_x * (  cos(a)/s )
          + new_y * ( -sin(a)/s )
          +         (  x0 )
    old_y = y0 + new_y*cos(a)/s + new_x*sin(a)/s
          = new_x * ( sin(a)/s )
          + new_y * ( cos(a)/s )
          +         ( y0 )
*/

typedef struct {
    HPGL_Pt x;
    HPGL_Pt y;
    HPGL_Pt t;
} HPGL_Trans;

static void define_rotation(HPGL_Trans *forward, HPGL_Trans *backward,
                            const HPGL_Pt center, const float angle, const float scale)
{
    const double  a = (double)angle * M_PI / 180.0;

    if (forward) {
        const double  c = cos(a) * scale;
        const double  s = sin(a) * scale;

        forward->x.x = c;
        forward->x.y = s;
        forward->t.x = -center.x * c - center.y * s;

        forward->y.x = -s;
        forward->y.y =  c;
        forward->t.y = -center.y * c + center.x * s;
    }

    if (backward) {
        const double  c = cos(a) / scale;
        const double  s = sin(a) / scale;

        backward->x.x = c;
        backward->x.y = -s;
        backward->t.x = center.x;

        backward->y.x = s;
        backward->y.y = c;
        backward->t.y = center.y;
    }
}

/* Rotate the source polygon by 'angle', scaling it by 'scale'.
   Returns the number of unique points in the resulting polygon,
   with (*polyptr)[points] = (*polyptr)[0].
*/
static int rotate_poly(HPGL_Pt **polyptr, const HPGL_Pt source[], int points,
                       HPGL_Trans *back, const float angle, const float scale)
{
    HPGL_Trans  trans;
    HPGL_Pt    *poly, center;
    int         count, i;

    if (polyptr)
        *polyptr = NULL;

    if (points < 1)
        return 0;

    /* 'points' includes only unique points. */
    if (source[0].x == source[points-1].x && source[0].y == source[points-1].y)
        points--;

    if (points < 1)
        return 0;

    poly = calloc(points + 1, sizeof (HPGL_Pt));
    if (!poly)
        return 0;

    /* We use the centroid of the polygon as the origin for the rotated polygon,
       to ensure we use sensible coordinates for the polygon.
       Note: we should probably use P1,P2 to adjust the origin? */
    center = centroid(source, points);

    define_rotation(&trans, back, center, angle, scale);

    poly[0].x = source[0].x * trans.x.x + source[0].y * trans.x.y + trans.t.x;
    poly[0].y = source[0].x * trans.y.x + source[0].y * trans.y.y + trans.t.y;
    count = 1;
    for (i = 1; i < points; i++) {
        const double  x = source[i].x;
        const double  y = source[i].y;

        poly[count].x = x * trans.x.x + y * trans.x.y + trans.t.x;
        poly[count].y = x * trans.y.x + y * trans.y.y + trans.t.y;

        if (poly[count].x != poly[count-1].x || poly[count].y != poly[count-1].y)
            count++;
    }

    poly[count] = poly[0];
    if (poly[0].x != poly[count-1].x || poly[0].y != poly[count-1].y)
        count++;

    if (polyptr)
        *polyptr = poly;
    else
        free(poly);

    return count;
}

/* Do Pen_action_to_tmpfile(), except applying transform to the point.
*/
static void do_pen_action(PlotCmd cmd, const HPGL_Pt point, int scaled, HPGL_Trans *transform)
{
    if (transform) {
        HPGL_Pt  p;
        p.x = point.x * transform->x.x + point.y * transform->x.y + transform->t.x;
        p.y = point.x * transform->y.x + point.y * transform->y.y + transform->t.y;
        Pen_action_to_tmpfile(cmd, &p, scaled);
    } else
        Pen_action_to_tmpfile(cmd, &point, scaled);
}

/* Return the axis-aligned bounding box for the specified polygon.
*/
static void axis_aligned_bounding_box(HPGL_Pt poly[], int points, HPGL_Pt *minptr, HPGL_Pt *maxptr)
{
    HPGL_Pt  lower = poly[0];
    HPGL_Pt  upper = poly[0];
    int      i;

    for (i = 1; i < points; i++) {
        if (lower.x > poly[i].x)
            lower.x = poly[i].x;

        if (upper.x < poly[i].x)
            upper.x = poly[i].x;

        if (lower.y > poly[i].y)
            lower.y = poly[i].y;

        if (upper.y < poly[i].y)
            upper.y = poly[i].y;
    }

    if (minptr)
        *minptr = lower;
    if (maxptr)
        *maxptr = upper;
}

/* octants:  +y
 *         |5 3 4|
 *       -x|2 0 1|+x
 *         |8 6 7|
*/

static inline int octant(const float x, const float y)
{
    return (x > 0.0 ? 1 : 0)
         + (x < 0.0 ? 2 : 0)
         + (y > 0.0 ? 3 : 0)
         + (y < 0.0 ? 6 : 0);
}

static const int octant_winding[9][9] = {
    { -0, -0, -0, -0, -0, -0, -0, -0, -0 }, /* -0: Start or end point at vertex */
    { -0,  0, 00, -2, -1, -3,  2,  1,  3 }, /* 00: On boundary line */
    { -0, 00,  0,  2,  3,  1, -2, -3, -1 }, /* +0: Intersection check needed */
    { -0,  2, -2,  0,  1, -1, 00,  3, -3 },
    { -0,  1, -3, -1,  0, -2,  3,  2, +0 },
    { -0,  3, -1,  1,  2,  0, -3, +0, -2 },
    { -0, -2,  2, 00, -3,  3,  0, -1,  1 },
    { -0, -1,  3, -3, -2, +0,  1,  0,  2 },
    { -0, -3,  1,  3, +0,  2, -1, -2,  0 }
};

typedef enum {
    INSIDE   = 0,
    BOUNDARY = 1,
    OUTSIDE  = 2
} inout;

/* Determine whether point 'p' is within the polygon.
*/
static inout inside(const HPGL_Pt  point[],
                    const int      points,
                    const HPGL_Pt  p,
                    const int      nonzero)
{
    int    winding = 0; /* +-8 per full turn */
    float  prev_x, prev_y, curr_x, curr_y, distance;
    int    prev_octant, curr_octant, i;

    curr_x = point[points-1].x - p.x;
    curr_y = point[points-1].y - p.y;
    curr_octant = octant(curr_x, curr_y);
    if (!curr_octant)
        return BOUNDARY;

    for (i = 0; i < points; i++) {

        prev_x = curr_x;
        prev_y = curr_y;
        prev_octant = curr_octant;

        curr_x = point[i].x - p.x;
        curr_y = point[i].y - p.y;
        curr_octant = octant(curr_x, curr_y);
        if (!curr_octant)
            return BOUNDARY;

        if (prev_octant == curr_octant)
            continue;

        /* Polygon edge passes through the point? */
        if ((prev_octant == 2 && curr_octant == 1) ||
            (prev_octant == 1 && curr_octant == 2) ||
            (prev_octant == 3 && curr_octant == 6) ||
            (prev_octant == 6 && curr_octant == 3))
            return BOUNDARY;

        if (prev_octant == 4 && curr_octant == 8) {
            /* +x,+y to -x,-y */
            const float dx = prev_x - curr_x,
                        dy = prev_y - curr_y;

            if (dx >= dy) {
                if (prev_x >= -curr_x)
                    distance = (curr_x * dy) / dx - curr_y;
                else
                    distance = (prev_x * dy) / dx - prev_y;
            } else {
                if (prev_y >= -curr_y)
                    distance = curr_x - (curr_y * dx) / dy;
                else
                    distance = prev_x - (prev_y * dx) / dy;
            }

            if (distance > 0.0f)
                winding += 4;
            else
            if (distance < 0)
                winding -= 4;
            else
                return BOUNDARY;

            continue;
        }

        if (prev_octant == 5 && curr_octant == 7) {
            /* -x,+y to +x,-y */
            const float dx = curr_x - prev_x,
                        dy = prev_y - curr_y;

            if (dx >= dy) {
                if (-prev_x >= curr_x)
                    distance = curr_y + (curr_x * dy) / dx;
                else
                    distance = prev_y + (prev_x * dy) / dx;
            } else {
                if (prev_y >= -curr_y)
                    distance = curr_x + (curr_y * dx) / dy;
                else
                    distance = prev_x + (prev_y * dx) / dy;
            }

            if (distance > 0.0f)
                winding += 4;
            else
            if (distance < 0.0f)
                winding -= 4;
            else
                return BOUNDARY;

            continue;
        }

        if (prev_octant == 7 && curr_octant == 5) {
            /* +x,-y to -x,+y */
            const float dx = prev_x - curr_x,
                        dy = curr_y - prev_y;
            if (dx >= dy) {
                if (prev_x >= -curr_x)
                    distance = (-curr_x * dy) / dx - curr_y;
                else
                    distance = (-prev_x * dy) / dx - prev_y;
            } else {
                if (-prev_y >= curr_y)
                    distance = (-curr_y * dx) / dy - curr_x;
                else
                    distance = (-prev_y * dx) / dy - prev_x;
            }

            if (distance > 0.0f)
                winding += 4;
            else
            if (distance < 0.0f)
                winding -= 4;
            else
                return BOUNDARY;

            continue;
        }

        if (prev_octant == 8 && curr_octant == 4) {
            /* -x,-y to +x,+y */
            const float dx = curr_x - prev_x,
                        dy = curr_y - prev_y;

            if (dx >= dy) {
                if (-prev_x >= curr_x)
                    distance = curr_y - (curr_x * dy) / dx;
                else
                    distance = prev_y - (prev_x * dy) / dx;
            } else {
                if (-prev_y >= curr_y)
                    distance = (curr_y * dx) / dy - curr_x;
                else
                    distance = (prev_y * dx) / dy - prev_x;
            }

            if (distance > 0.0f)
                winding += 4;
            else
            if (distance < 0.0f)
                winding -= 4;
            else
                return BOUNDARY;

            continue;
        }

        winding += octant_winding[prev_octant][curr_octant];
    }

    /* Adjust for polygon orientation. */
    if (winding < 0)
        winding = -winding;

    /* Nonzero mode? */
    if (nonzero) {
        if (winding < 4)
            return OUTSIDE;
        else
            return INSIDE;
    }

    /* Odd-even mode; limit to two full turns. */
    winding &= 15;
    if (winding < 4 || winding > 11)
        return OUTSIDE;
    else
        return INSIDE;
}

/* Insert a float into a sorted array, in increasing value.
*/
static int insert_float(float list[], int count, float value)
{
    int  i = 0;

    while (i < count && list[i] < value)
        i++;

    if (i >= count) {
        list[count] = value;
        return count + 1;
    }

    if (list[i] == value)
        return count;

    memmove(&(list[i+1]), &(list[i]), (size_t)(count - i) * sizeof list[0]);

    list[i] = value;
    return count + 1;
}

/* Reverse an array of floats.
*/
static void reverse_float(float list[], int count)
{
    float *p = &(list[0]);
    float *q = &(list[count-1]);

    while (p < q) {
        const float old_p = *p;
        const float old_q = *q;
        *(p++) = old_q;
        *(q--) = old_p;
    }
}

/* Calculate where line 'p0','p1' intersects y='y', and insert that x coordinate into 'list'.
*/
static int insert_x(float list[], int count, const HPGL_Pt p0, const HPGL_Pt p1, const float y)
{
    const float  p = (y - p0.y) / (p1.y - p0.y);
    const float  q = 1.0f - p;
    return insert_float(list, count, (q*p0.x) + (p*p1.x));
}

/* Calculate where line 'p0','p1' intersects x='x', and insert that y coordinate into 'list'.
*/
static int insert_y(float list[], int count, const HPGL_Pt p0, const HPGL_Pt p1, const float x)
{
    const float  p = (x - p0.x) / (p1.x - p0.x);
    const float  q = 1.0f - p;
    return insert_float(list, count, (q*p0.y) + (p*p1.y));
}

/* Hatch and crosshatch fill.
*/
static void do_fill_lines(HPGL_Pt polygon[],
                          int     polygon_points,
                          int     scale_flag,
                          int     filltype,   /* 1 to 4 */
                          int     bidir,
                          int     nonzero,
                          float   angle,      /* degrees */
                          float   scale)
{
    HPGL_Pt    *poly = NULL;
    HPGL_Pt     lower, upper, middle;
    HPGL_Trans  back;
    float      *coord;
    int         poly_points, coord_count, cmin, cmax, c, i;

    if (polygon_points < 3)
        return;

    poly_points = rotate_poly(&poly, polygon, polygon_points, &back, angle, scale);
    if (poly_points < 3) {
        free(poly);
        return;
    }

    /* Note: poly[0] == poly[poly_points]; i.e. poly_points is the number of vertices. */

    /* At most, we can have poly_points intersections. Add two for safety. */
    coord = calloc(poly_points + 2, sizeof coord[0]);
    if (!coord) {
        free(poly);
        return;
    }

    axis_aligned_bounding_box(poly, poly_points, &lower, &upper);

    if (filltype >= 1 && filltype <= 4) {
        /* Fill with lines along the X axis, at unit intervals. */
        cmin = (int)lower.y - 1;
        cmax = (int)upper.y + 1;
        for (c = cmin; c <= cmax; c++) {
            middle.y = c;
            coord_count = 0;

            for (i = 0; i < poly_points; i++) {
                if (poly[i].y == middle.y)
                    coord_count = insert_float(coord, coord_count, poly[i].x);
                else
                if (poly[i].y < middle.y && poly[i+1].y > middle.y)
                    coord_count = insert_x(coord, coord_count, poly[i], poly[i+1], middle.y);
                else
                if (poly[i].y > middle.y && poly[i+1].y < middle.y)
                    coord_count = insert_x(coord, coord_count, poly[i], poly[i+1], middle.y);
            }

            if (coord_count < 2)
                continue;

            if ((bidir) && (c & 1))
                reverse_float(coord, coord_count);

            for (i = 1; i < coord_count; i++) {
                middle.x = (0.5f * coord[i-1]) + (0.5f * coord[i]);
                if (inside(poly, poly_points, middle, nonzero) == INSIDE) {
                    middle.x = coord[i-1];
                    do_pen_action(MOVE_TO, middle, scale_flag, &back);
                    middle.x = coord[i];
                    do_pen_action(DRAW_TO, middle, scale_flag, &back);
                }
            }
        }
    }

    if (filltype == 4) {
        /* Fill with lines along the Y axis, at unit intervals. */
        cmin = (int)lower.x - 1;
        cmax = (int)upper.x + 1;
        for (c = cmin; c <= cmax; c++) {
            middle.x = c;
            coord_count = 0;

            for (i = 0; i < poly_points; i++) {
                if (poly[i].x == middle.x)
                    coord_count = insert_float(coord, coord_count, poly[i].y);
                else
                if (poly[i].x < middle.x && poly[i+1].x > middle.x)
                    coord_count = insert_y(coord, coord_count, poly[i], poly[i+1], middle.x);
                else
                if (poly[i].x > middle.x && poly[i+1].x < middle.x)
                    coord_count = insert_y(coord, coord_count, poly[i], poly[i+1], middle.x);
            }

            if (coord_count < 2)
                continue;

            if ((bidir) && (c & 1))
                reverse_float(coord, coord_count);

            for (i = 1; i < coord_count; i++) {
                middle.y = (0.5f * coord[i-1]) + (0.5f * coord[i]);
                if (inside(poly, poly_points, middle, nonzero) == INSIDE) {
                    middle.y = coord[i-1];
                    do_pen_action(MOVE_TO, middle, scale_flag, &back);
                    middle.y = coord[i];
                    do_pen_action(DRAW_TO, middle, scale_flag, &back);
                }
            }
        }
    }

    free(coord);
    free(poly);
}

void fill(HPGL_Pt polygon[],
          int     numpoints,
          HPGL_Pt point1,
          HPGL_Pt point2,
          int     scale_flag,
          int     filltype,
          float   spacing,
          float   hatchangle,
          float   curwidth,
          int     curdpi,
          int     nzfill,
          int     bifill)
{
    if (numpoints < 3)
        return;

    switch (filltype) {

    case 1:
    case 2:
        do_fill_lines(polygon, numpoints, scale_flag, filltype, bifill, nzfill, hatchangle, 1.0f / curwidth);
        return;

    case 3:
    case 4:
        do_fill_lines(polygon, numpoints, scale_flag, filltype, bifill, nzfill, hatchangle, 1.0f / spacing);
        return;
    }
}
